/* board.cc Copyright 2001 Robert Hilchie */

#include <stdlib.h>
#include <stdio.h>
#include "board.h"

static char movenames[] = "UDLR";

Board::Board()
{
  // Start printing in column 0
  column = 0;
}

// Set to the solved configuration
void Board::home()
{
  unsigned x, y, t;

  for (y=0; y<ROWS; y++)
  {
    for (x=0; x<COLS; x++)
    {
      t = 4*y + x + 1;

      // The blank tile goes in the lower-right position
      if (t==TILES)
        t = BLANKTILE;

      tiles[y][x] = t;
      tilex[t] = x;
      tiley[t] = y;
    }
  }
}

#ifdef DEBUG
void Board::output()
{
  unsigned x, y;

  printf("\n");
  for (y=0; y<ROWS; y++)
  {
    for (x=0; x<COLS; x++)
    {
      if (tiles[y][x]==0)
        printf("   ");
      else
        printf("%3u", tiles[y][x]);
    }
    printf("\n");
  }
}
#endif

// Input a configuration
bool Board::input()
{
  unsigned x, y, t;

  for (y=0; y<TILES; y++)
    tiley[y] = ROWS;

  for (y=0; y<ROWS; y++)
    for (x=0; x<COLS; x++)
    {
      if (scanf("%u", &t) != 1)
      {
        printf("Error: please enter only numbers.\n");
        return false;
      }
      if (t >= TILES)
      {
        printf("Error: tiles must be numbered from 0 to %u.\n", TILES-1);
        return false;
      }
      if (tiley[t] != ROWS)
      {
        printf("Error: the tile %u is repeated.\n", t);
        return false;
      }
      tiles[y][x] = t;
      tilex[t] = x;
      tiley[t] = y;
    }
  return validate();
}

// Determine whether the current configuration can be solved
bool Board::validate()
{
  unsigned pos[TILES];
  unsigned y, x, i, swaps;

  // Copy the tiles into the linear array 'pos', such that
  // pos[t] is the tile that is in the home position of tile t
  i=1;
  for (y=0; y<ROWS; y++)
  {
    for (x=0; x<COLS; x++)
    {
      pos[i++] = tiles[y][x];
      if (i==TILES)
        i=0;
    }
  }

  // Sort the array 'pos', counting the number of swaps
  swaps = 0;
  for (x=0; x<TILES-1; x++)
  {
    if (pos[x] != x)
    {
      for (y=x+1; y<TILES; y++)
      {
        if (pos[y] == x)
          break;
      }
      i = pos[x];
      pos[x] = pos[y];
      pos[y] = i;
      swaps++;
    }
  }

  // If the permutation of tiles (even or odd) is inconsistent with
  // the position of the blank tile, return false
  if (((swaps + tilex[BLANKTILE] + tiley[BLANKTILE]) & 1) != 0)
  {
    printf("The puzzle can not be solved in this configuration.\n");
    return false;
  }

  return true;
}

// Move a tile up into the blank's position. If the tile moved is part
// of the current pattern or no pattern is involved, then count = true.
// Otherwise, count = false. If the move is not possible, false is
// returned, otherwise true.
bool Board::moveup(bool &count)
{
  unsigned x, y, t;

  y = tiley[BLANKTILE];
  if (y==(ROWS-1))
    return false;

  x = tilex[BLANKTILE];
  t = tiles[y+1][x];
  tiles[y+1][x] = BLANKTILE;
  tiles[y][x] = t;
  tiley[BLANKTILE]++;

  count = false;
  if (t != DONTCARETILE)
  {
    tiley[t]--;
    count = true;
  }
  lastmoved = t;
  return true;
}

// Move a tile down into the blank's position. If the tile moved is part
// of the current pattern or no pattern is involved, then count = true.
// Otherwise, count = false. If the move is not possible, false is
// returned, otherwise true.
bool Board::movedown(bool &count)
{
  unsigned x, y, t;

  y = tiley[BLANKTILE];
  if (y==0)
    return false;

  x = tilex[BLANKTILE];
  t = tiles[y-1][x];
  tiles[y-1][x] = BLANKTILE;
  tiles[y][x] = t;
  tiley[BLANKTILE]--;

  count = false;
  if (t != DONTCARETILE)
  {
    tiley[t]++;
    count = true;
  }
  lastmoved = t;
  return true;
}

// Move a tile left into the blank's position. If the tile moved is part
// of the current pattern or no pattern is involved, then count = true.
// Otherwise, count = false. If the move is not possible, false is
// returned, otherwise true.
bool Board::moveleft(bool &count)
{
  unsigned x, y, t;

  x = tilex[BLANKTILE];
  if (x==(COLS-1))
    return false;

  y = tiley[BLANKTILE];
  t = tiles[y][x+1];
  tiles[y][x+1] = BLANKTILE;
  tiles[y][x] = t;
  tilex[BLANKTILE]++;

  count = false;
  if (t != DONTCARETILE)
  {
    tilex[t]--;
    count = true;
  }
  lastmoved = t;
  return true;
}

// Move a tile right into the blank's position. If the tile moved is part
// of the current pattern or no pattern is involved, then count = true.
// Otherwise, count = false. If the move is not possible, false is
// returned, otherwise true.
bool Board::moveright(bool &count)
{
  unsigned x, y, t;

  x = tilex[BLANKTILE];
  if (x==0)
    return false;

  y = tiley[BLANKTILE];
  t = tiles[y][x-1];
  tiles[y][x-1] = BLANKTILE;
  tiles[y][x] = t;
  tilex[BLANKTILE]--;

  count = false;
  if (t != DONTCARETILE)
  {
    tilex[t]++;
    count = true;
  }
  lastmoved = t;
  return true;
}

// Do a moved indexed by the number 'm'
bool Board::move(unsigned m, bool &count)
{
  switch(m)
  {
    case 0:
      return moveup(count);

    case 1:
      return movedown(count);

    case 2:
      return moveleft(count);
  }
  return moveright(count);
}

// Do a move and print it out
void Board::moveandprint(unsigned m)
{
  bool count;

  move(m, count);
  printf("%2u%c ", lastmoved, movenames[m]);

  column++;
  if (column == 15)
  {
    column = 0;
    printf("\n");
  }
}

// Given a pattern, compute the index of the current configuration
Index Board::index(Pattern &pattern)
{
  unsigned pos[TILES];
  unsigned i, j, mult;
  Index total;

  // Compute the position number for each tile in the pattern
  for (i=0; i<pattern.size(); i++)
    pos[i] = COLS*tiley[pattern.itile(i)] + tilex[pattern.itile(i)];

  // Combine the position numbers into a single index
  mult = TILES; // Initially, all positions available
  total = 0;
  for (i=0; i<(pattern.size()-1); i++)
  {
    total += pos[i]; // Accumulate position of 'i'th tile

    // Adjust position numbers of subsequent tiles because fewer
    // positions are available
    for (j=i+1; j<pattern.size(); j++)
      if (pos[i] < pos[j])
        pos[j]--;

    mult--;  // One fewer position is available
    total *= mult;
  }
  total += pos[i]; //Accumulate final position number

  return total;
}

// Given a pattern and index, create the corresponding configuration
void Board::unindex(Index idx, Pattern &pattern)
{
  unsigned pos[TILES];
  int i, j;
  unsigned div;

  // Break apart the index into position numbers
  div = TILES - pattern.size();
  for (i=pattern.size()-1; i>=0; i--)
  {
    div++;
    pos[i] = idx % div;
    idx /= div;
  }

  // Compute true, unadjusted position numbers
  for (i=pattern.size()-1; i>=0; i--)
    for (j=i+1; j<(int)pattern.size(); j++)
      if (pos[i] <= pos[j])
        pos[j]++;

  // Set all tiles to DONTCARETILE initially
  for (i=0; i<ROWS; i++)
    for (j=0; j<COLS; j++)
      tiles[i][j] = DONTCARETILE;

  // Lay out the tiles of the pattern according to their position numbers
  for (i=0; i<(int)pattern.size(); i++)
  {
    tilex[pattern.itile(i)] = pos[i]&3;
    tiley[pattern.itile(i)] = pos[i]>>2;
    tiles[tiley[pattern.itile(i)]][tilex[pattern.itile(i)]] = pattern.itile(i);
  }
}

// Swap the tiles t1 and t2
void Board::swaptiles(Tile t1, Tile t2)
{
  unsigned sw;

  sw = tilex[t1];
  tilex[t1] = tilex[t2];
  tilex[t2] = sw;

  sw = tiley[t1];
  tiley[t1] = tiley[t2];
  tiley[t2] = sw;

  tiles[tiley[t1]][tilex[t1]] = t1;
  tiles[tiley[t2]][tilex[t2]] = t2;
}

// Swap the tiles at positions (x1, y1) and (x2, y2)
void Board::swappos(unsigned y1, unsigned x1, unsigned y2, unsigned x2)
{
  swaptiles(tiles[y1][x1], tiles[y2][x2]);
}

// Conjugate the current configuration with a horizontal reflection
void Board::hreflect()
{
  unsigned y;

  swaptiles(1, 4);
  swaptiles(2, 3);
  swaptiles(5, 8);
  swaptiles(6, 7);
  swaptiles(9, 12);
  swaptiles(10, 11);
  swaptiles(13, 15);
  for (y=0; y<ROWS; y++)
  {
    swappos(y, 0, y, 3);
    swappos(y, 1, y, 2);
  }
}

// Conjugate the current configuration with a vertical reflection
void Board::vreflect()
{
  unsigned x;

  swaptiles(1, 13);
  swaptiles(5, 9);
  swaptiles(2, 14);
  swaptiles(6, 10);
  swaptiles(3, 15);
  swaptiles(7, 11);
  swaptiles(4, 12);
  for (x=0; x<ROWS; x++)
  {
    swappos(0, x, 3, x);
    swappos(1, x, 2, x);
  }
}
