/* idasearch.cc Copyright 2001 Robert Hilchie */

#include <stdio.h>
#include <time.h>
#include "idasearch.h"

// Each move horizontally reflected
static unsigned hrefmove[] = {0, 1, 3, 2};

// Each move vertically reflected
static unsigned vrefmove[] = {1, 0, 2, 3};

extern unsigned invmove[];

// Do a recursive heuristically guided depth-first search
// level  - the number of moves this configuration is assumed
//          to be away from the home configuration
// forbid - the inverse of the previous move, so we don't
//          immediately follow a move by its inverse
// i1, i2, i3, i4 - encodes the configuration of the puzzle as
//                  indices into the pattern database
bool IDAsearch::dfs(unsigned level, unsigned forbid,
                    Index i1, Index i2, Index i3, Index i4)
{
  unsigned move;
  unsigned char heuristic;
  Index *mt1, *mt2, *mt3, *mt4;

  nodecount++;

  // Estimate number of moves to home configuration
  heuristic = dbase4[i1] + dbase4[i2] + dbase4[i3] + dbase3[i4];

  // If too high, bail out
  if (heuristic > (unsigned char)level)
    return false;

  // Solution found!
  if (level == 0)
    return true;

  level--;

  // Get the applicable rows of the move tables
  mt1 = movetable4[i1];
  mt2 = movetable4[i2];
  mt3 = movetable4[i3];
  mt4 = movetable3[i4];

  // For each possible move
  for (move=0; move<4; move++)
  {
    // If it's not the inverse of the previous move
    if (move != forbid)
    {
      // If the move is valid in this configuration
      if (*mt1 != NOMOVE)
      {
        // Apply the move and recursively search
        if (dfs(level, invmove[move], *mt1, mt2[hrefmove[move]],
                mt3[vrefmove[move]], *mt4))
        {
          // Solution found. This move is part of the solution.
          solution[level] = move;
          return true;
        }
      }
    }
    mt1++;
    mt4++;
  }
  // Solution not found in this branch of search
  return false;
}

// Perform the iterative deepening A* search
void IDAsearch::search(Pattern &p4, Pattern &p3, Board &b)
{
  Index i1, i2, i3, i4;
  unsigned depth;
  unsigned total;
  bool done;
  clock_t t1, t2;

  movetable4 = p4.getmovetable();
  dbase4 = p4.getdbase();

  movetable3 = p3.getmovetable();
  dbase3 = p3.getdbase();

  // Compute index corresponding to the configuration of tiles
  // 1, 2, 5, 6 and blank.
  i1 = b.index(p4);

  // Compute index corresponding to the configuration of tiles
  // 11, 12, 15 and blank.
  i4 = b.index(p3);

  // Compute index corresponding to the configuration of tiles
  // 3, 4, 7, 8 and blank by conjugating the puzzle with a
  // horizontal reflection.
  b.hreflect();
  i2 = b.index(p4);

  // Undo the reflection
  b.hreflect();

  // Compute index corresponding to the configuration of tiles
  // 9, 10, 13, 14 and blank by conjugating the puzzle with a
  // vertical reflection.
  b.vreflect();
  i3 = b.index(p4);

  // Undo the reflection
  b.vreflect();

  // The first depth-first search will have a depth equal
  // to the initial estimated number of moves
  depth = dbase4[i1]+dbase4[i2]+dbase4[i3]+dbase3[i4];

  printf("Initial estimate: %u + %u + %u + %u = %u moves\n", 
   dbase4[i1], dbase4[i2], dbase4[i3], dbase3[i4], depth);

  t1 = clock();
  total = 0;

  // Do depth-first searches of increasing depth until a solution
  // is found
  do
  {
    printf("%4u moves: ", depth);
    nodecount = 0;

    done = dfs(depth, 99, i1, i2, i3, i4);

    printf("%10u nodes\n", nodecount);
    total += nodecount;
    nodecount = 0;

    if (done)
      break;

    depth += 2;
  }
  while (depth < MAXDEPTH);

  t2 = clock();

  printf("%u nodes total\n", total);
  printf("%f seconds\n", ((double)(t2-t1))/CLOCKS_PER_SEC);
  if (t1 != t2)
    printf("%f nodes per second\n", ((double)total)/(t2-t1)*CLOCKS_PER_SEC);

  if (depth < MAXDEPTH)
  {
    // Print out the solution
    do
    {
      depth--;
      b.moveandprint(solution[depth]);
    }
    while (depth>0);
  }
  else
    printf("No solution found.\n");
}

// List of tiles for each of the patterns
static Tile pattern4[] = {1, 2, 5, 6, 0};
static Tile pattern3[] = {11, 12, 15, 0};

int main(int argc, char **argv)
{
  Board b;

  printf("IDA15 version 1.0 - An optimal solver for the fifteen puzzle\n");
  printf("Copyright 2001 Robert Hilchie\n");
  printf("Enter the initial configuration now:\n");
  if (b.input())
  {
    Pattern pat3(4, pattern3);
    Pattern pat4(5, pattern4);

    printf("Generating databases. Please wait.\n");
    if (pat3.makemovetable() &&
        pat3.makedbase() &&
        pat4.makemovetable() &&
        pat4.makedbase())
    {
      IDAsearch ida;

      ida.search(pat4, pat3, b);
    }
  }
  return 1;
}