/* pattern.cc Copyright 2001 Robert Hilchie */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "board.h"

// Unassigned entry in the pattern database
#define DBEMPTY 255

// The inverse of each move
unsigned invmove[] = {1, 0, 3, 2};

// tls is an array of tiles in the pattern, tc is the size of tls
Pattern::Pattern(unsigned tc, Tile *tls)
{
  unsigned p;

  tilecount = tc;
  tiles = tls;

  // Compute the size of the pattern datbase
  tablesize = 1;
  for (p=TILES-tc+1; p<=TILES; p++)
    tablesize *= p;

  movetable = NULL;
  counttable = NULL;
  dbase = NULL;
}

Pattern::~Pattern()
{
  if (movetable != NULL)
    delete movetable;

  if (counttable != NULL)
    delete counttable;

  if (dbase != NULL)
    delete dbase;
}

// Create the move table
bool Pattern::makemovetable()
{
  Index i;
  unsigned m;
  Board b;
  bool count;

  movetable = new Movetable[tablesize];
  if (movetable==NULL)
  {
    printf("Failed to allocate memory for move table.\n");
    return false;
  }

  counttable = new Counttable[tablesize];
  if (counttable==NULL)
  {
    printf("Failed to allocate memory for count table.\n");
    return false;
  }

  // For each index in the move table
  for (i=0; i<tablesize; i++)
  {
    // Create the corresponding configuration of tiles
    b.unindex(i, *this);

    // Perform each possible move on the configuration
    for (m=0; m<4; m++)
    {
      movetable[i][m] = NOMOVE;
      if (b.move(m, count))
      {
        // Compute new index and count, store in tables
        movetable[i][m] = b.index(*this);
        counttable[i][m] = count;

        // Undo the move
        b.move(invmove[m], count);
      }
    }
  }
  return true;
}

// Do a breadth-first search pass on the pattern database
// In this pass, only database entries that are 'level' moves
// from the home position are assigned. Only entries that are
// 'baselevel' or higher are examined as parent nodes.
bool Pattern::bfspass(unsigned level, unsigned baselevel)
{
  Index i, newindex;
  unsigned j;
  bool bret;

  bret = false;

  // For every database entry
  for (i=0; i<tablesize; i++)
  {
    // If it's already assigned
    if (dbase[i] != DBEMPTY)
    {
      // If it qualifies as a possible parent node
      if (dbase[i] >= baselevel)
      {
        // For each possible move
        for (j=0; j<4; j++)
        {
          // Apply the move (if valid) to the parent index
          newindex = movetable[i][j];
          if (newindex != NOMOVE)
	  {
            // If the new entry is not yet assigned
            if (dbase[newindex] == DBEMPTY)
            {
              // Compute the number of moves from the home position.
              // If it belongs to this pass...
              if (((unsigned)(dbase[i] + counttable[i][j])) == level)
              {
                // Assing the new database entry
                dbase[newindex] = level;
                bret = true;
                dbcount++;
                if (dbcount == tablesize)
                  return true;
              }
            }
          }
        }
      }
    }
  }
  return bret;
}

// Create the pattern database
bool Pattern::makedbase()
{
  Board b;
  unsigned level;

  dbase = new unsigned char[tablesize];
  if (dbase==NULL)
  {
    printf("Failed to allocate memory for data base.\n");
    return false;
  }

  // Mark all entries as unassigned
  memset(dbase, DBEMPTY, tablesize);

  // Assign 0 to the entry corresponding to the home configuration
  b.home();
  dbase[b.index(*this)] = 0;

  // In the following loop, 'level' represents the minimum number of 
  // moves required to set the pattern into its home position (counting
  // only moves affecting the non-blank tiles of the pattern). We want
  // to make sure that all database entries equal to 'level' are assigned
  // before assigning entries 'level'+1 and above.
  level = 0;
  dbcount = 1;
  do
  {
    // Do as many bfs passes on the current level as required
    // until this level is exhausted
    while(bfspass(level, level))
      if (dbcount == tablesize)
        return true;

    // Do a single pass on the current level to start the
    // next level
    level++;
    bfspass(level, level-1);
  }
  while (dbcount < tablesize);

  delete counttable;
  counttable = NULL;

  return true;
}